#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include "heap.h"
#include "traverse.h"

typedef struct File {
   char* path;
   struct timespec mtime;
} File;

int compare_times(struct timespec a, struct timespec b) {
   if (a.tv_sec == b.tv_sec) {
      if (a.tv_nsec > b.tv_nsec) {
	 return 1;
      } else if (a.tv_nsec < b.tv_nsec) {
	 return -1;
      } else {
	 return 0;
      }
   } else {
      if (a.tv_sec > b.tv_sec) {
	 return 1;
      } else if (a.tv_sec < b.tv_sec) {
	 return -1;
      } else {
	 return 0;
      }
   }
}

int compare_oldest(void* a, void* b) {
   struct timespec time_a = ((File*) a)->mtime;
   struct timespec time_b = ((File*) b)->mtime;
   return compare_times(time_a, time_b);
}

int compare_newest(void* a, void* b) {
   return -compare_oldest(a, b);
}

static bool newest_flag = true;

bool handler(void* handle, char* path, struct stat* statbuf) {
   Heap* heap = (Heap*) handle;

   if (!S_ISREG(statbuf->st_mode)) return true;

   bool filled = get_heap_size(heap) == get_heap_maxsize(heap);
   if (filled) {
      bool insert = false;
      struct timespec current_max = ((File*)get_max_from_heap(heap))->mtime;
      if (newest_flag) {
	 insert = compare_times(statbuf->st_mtim, current_max) > 0;
      } else {
	 insert = compare_times(statbuf->st_mtim, current_max) < 0;
      }
      if (!insert) return true; 
      /* remove current max from the heap, i.e.
         if newest_flag: remove oldest file from the heap to
	                 keep the newest files on the heap
         if !newest_flag: remove newest file from the heap to
	                  keep the oldest files on the heap
      */
      File* file = extract_max_from_heap(heap);
      free(file->path); free(file);
   }
   File* file = malloc(sizeof(File));
   file->path = strdup(path); file->mtime = statbuf->st_mtim;
   insert_into_heap(heap, file);
   return true;
}

int main(int argc, char** argv) {
   char* cmdname = *argv;
   char usage[] = "Usage: %s [-c count] [-d] [-o] [dirpath]\n";
   size_t count = 1;
   bool date_flag = false;

   int option;
   while ((option = getopt(argc, argv, "c:do")) != -1) {
      switch (option) {
	 case 'c':
	    {
	       char* endptr = 0;
	       unsigned long val = strtoul(optarg, &endptr, 0);
	       if (*endptr || count == 0) {
		  fprintf(stderr, "%s: invalid value given to -c: %s\n",
		     cmdname, optarg);
		  fprintf(stderr, usage, cmdname); exit(1);
	       }
	       count = val;
	    }
	    break;
	 case 'd':
	    date_flag = true;
	    break;
	 case 'o':
	    newest_flag = false;
	    break;
	 default:
	    fprintf(stderr, usage, cmdname); exit(1);
	    break;
      }
   }

   argv += optind; argc -= optind;
   if (argc > 1) {
      fprintf(stderr, usage, cmdname); exit(1);
   }
   char* path = ".";
   if (argc > 0) path = argv[0];

   Heap* heap;
   if (newest_flag) {
      heap = create_heap(count, compare_newest);
   } else {
      heap = create_heap(count, compare_oldest);
   }
   if (!traverse(path, heap, handler)) {
      perror(path); exit(1);
   }

   File* file;
   while ((file = extract_max_from_heap(heap))) {
      if (date_flag) {
	 char* timestr = ctime(&file->mtime.tv_sec);
	 char* cp = strchr(timestr, '\n');
	 if (cp) *cp = 0;
	 printf("%s %s\n", timestr, file->path);
      } else {
	 printf("%s\n", file->path);
      }
      free(file->path); free(file);
   }
   destroy_heap(heap);
}
