#ifndef HEAP_H
#define HEAP_H

typedef struct Heap Heap;
typedef int (*Compare)(void*, void*);

/* create a heap with
   - maximal maxsize objects and
   - comparison function compare
   complexity: O(1)
*/
Heap* create_heap(size_t maxsize, Compare compare);

/* return current number of objects in heap;
   complexity: O(1) */
size_t get_heap_size(Heap* heap);

/* return maximal number of objects in heap;
   complexity: O(1) */
size_t get_heap_maxsize(Heap* heap);

/* insert another object into the heap;
   precondition: get_heap_size(heap) < get_heap_maxsize(heap);
   complexity: O(log(n)) */
void insert_into_heap(Heap* heap, void* object);

/* return the maximum object but do not remove it from the heap;
   returns 0 if there is no object in heap;
   complexity: O(1) */
void* get_max_from_heap(Heap* heap);

/* remove the maximum object from the heap and return it;
   0 is returned if the heap is empty;
   complexity: O(log(n)) */
void* extract_max_from_heap(Heap* heap);

/* free all resources associated with the given heap */
void destroy_heap(Heap* heap);

#endif
