#include <dirent.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include "traverse.h"

static bool rtraverse(int dirfd, char* dirpath, size_t dirpathlen,
      char* filename, void* handle, TraverseHandler handler);

static bool rscandir(int dirfd, char* dirpath, size_t dirpathlen,
      void* handle, TraverseHandler handler) {
   DIR* dir = fdopendir(dirfd);
   if (!dir) return false;
   struct dirent* entry;
   while ((entry = readdir(dir))) {
      if (entry->d_name[0] == '.' &&
	    (entry->d_name[1] == 0 ||
	       (entry->d_name[1] == '.' && entry->d_name[2] == 0))) {
	 /* skip '.' and '..' */
	 continue;
      }
      if (!rtraverse(dirfd, dirpath, dirpathlen, entry->d_name,
	    handle, handler)) {
	 closedir(dir);
	 return false;
      }
   }
   closedir(dir);
   return true;
}

static bool rtraverse(int dirfd, char* dirpath, size_t dirpathlen,
      char* filename,
      void* handle, TraverseHandler handler) {
   struct stat buf;
   if (fstatat(dirfd, filename, &buf, AT_SYMLINK_NOFOLLOW) < 0) return false;

   size_t pathlen = dirpathlen + strlen(filename) + 1;
   char* path = malloc(pathlen + 1);
   strcpy(path, dirpath);
   path[dirpathlen] = '/'; path[dirpathlen+1] = 0;
   strcat(path + dirpathlen + 1, filename);

   if (!handler(handle, path, &buf)) {
      free(path); return false;
   }
   if (!S_ISDIR(buf.st_mode)) {
      free(path); return true;
   }

   /* scan directory */
   int fd = openat(dirfd, filename, O_RDONLY);
   if (fd < 0) {
      free(path); return false;
   }
   bool ok = rscandir(fd, path, pathlen, handle, handler);
   close(fd); free(path);
   return ok;
}

bool traverse(char* path, void* handle, TraverseHandler handler) {
   int fd = open(path, O_RDONLY);
   if (fd < 0) return false;
   struct stat buf;
   if (fstat(fd, &buf) < 0) return false;
   bool rval = handler(handle, path, &buf);
   if (rval && S_ISDIR(buf.st_mode)) {
      rscandir(fd, path, strlen(path), handle, handler);
   }
   close(fd);
   return rval;
}
